from dataclasses import dataclass
from typing import List, Optional
from px_settings.contrib.django import settings as s


__all__ = 'Settings', 'settings'


@s('PX_ACCESS_TOKENS')
@dataclass
class Settings:
    AUTOLOAD_MODULE: str = 'access_scopes'
    REGISTRIES: Optional[List[str]] = None
    AGGREGATES: Optional[List[str]] = None


settings = Settings()



from typing import List, Union

from px_access_scopes.utils import flatten_aggregates

from ..scopes import ScopeRegistry
from ..aggregates import Aggregate, Aggregates
from .permissions import generate_permissions
from .groups import generate_group


__all__ = 'generate_all',


def generate_all(
    registries: List[ScopeRegistry],
    aggregates: List[Union[Aggregate, Aggregates]],
):
    for registry in registries:
        generate_permissions(registry)

    aggregates = flatten_aggregates(aggregates)

    for aggregate in aggregates:
        generate_group(aggregate)



class HasCallableMixin:
    def has(self, *a, **k) -> bool:
        raise NotImplemented('Implement `has` method. It\'s mandatory.')

    def __call__(self, *args, **kwargs) -> bool:
        return self.has(*args, **kwargs)



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy
from django.dispatch import receiver
from django.db.models.signals import post_migrate


__all__ = ('AccessScopesConfig',)


class AccessScopesConfig(AppConfig):
    name = 'px_access_scopes.contrib.drf'
    label = 'pxd_access_scopes_drf'
    verbose_name = pgettext_lazy('pxd_access_scopes', 'Access scopes: DRF')



from typing import Optional, Sequence, Set, TYPE_CHECKING

from .scopes import ScopeDomain

if TYPE_CHECKING:
    from django.contrib.auth.models import PermissionsMixin


__all__ = 'user_checker',


def user_checker(
    check_scopes: Set[ScopeDomain],
    user: Sequence['PermissionsMixin'] = None,
    obj: Optional[object] = None,
    **kwargs
) -> bool:
    """User scopes checker."""

    return user.has_perms(check_scopes, obj=obj)



from typing import Callable, Sequence, Set, Tuple
from functools import lru_cache

from ..scopes import ScopeRegistry, ScopeDomain
from .checkers import registry_hierarchy_lookup


__all__ = 'MultiregistryHierarchyLookup',


class MultiregistryHierarchyLookup:
    registries: Sequence[ScopeRegistry]
    descendands: Sequence[Tuple[ScopeRegistry, Set[ScopeDomain]]]
    lookup: Callable

    def __init__(
        self,
        registries: Sequence[ScopeRegistry] = [],
        hierarchy_lookup: Callable = registry_hierarchy_lookup,
    ):
        self.registries = registries
        self.descendands = []
        self.lookup = hierarchy_lookup

    def update_descendants(self):
        self.descendands = [
            (registry, registry.lookup_descendants())
            for registry in self.registries
        ]

    @lru_cache
    def get_registry(self, scope: ScopeDomain, first: bool = True):
        for registry, scopes in self.descendands:
            if scope in scopes:
                return registry

        if first:
            self.update_descendants()
            self.get_registry(scope, first=False)

    def __call__(self, scope: ScopeDomain, **kwargs) -> Set[ScopeDomain]:
        registry = self.get_registry(scope)

        return self.lookup(scope=scope, registry=registry, **kwargs)



from typing import Callable, Optional, Sequence, Set, Union
from functools import reduce

from ..scopes import ScopeRegistry, ScopeDomain
from ..aggregates import Aggregate
from ..utils import to_set, ScopeType
from .base import ScopesChecker


__all__ = (
    'run_checkers',
    'scopes_checker',
    'aggregates_checker',
    'registry_hierarchy_lookup',
    'domain_path_hierarchy_lookup',
    'HierarchyChecker',
)

ScopesCheckerType = Union['ScopesChecker', Callable[..., bool]]


def run_checkers(
    check_scopes: Set[ScopeDomain],
    checkers: Sequence['ScopesCheckerType'],
    kwargs: dict = {}
) -> bool:
    """Checkers runner function."""

    for checker in checkers:
        if checker(check_scopes, **kwargs):
            return True

    return False


def scopes_checker(
    check_scopes: Set[ScopeDomain],
    scopes: Sequence[ScopeType] = (),
    **kwargs
) -> bool:
    """Simple scopes to scopes checker."""

    return False if not scopes else len(to_set(scopes) & check_scopes) > 0


def aggregates_checker(
    check_scopes: Set[ScopeDomain],
    aggregates: Sequence[Aggregate] = (),
    **kwargs
) -> bool:
    """Checker for scopes in access aggregates."""

    return False if not aggregates else any(
        scopes_checker(check_scopes, scopes=aggregate)
        for aggregate in aggregates
    )


def registry_hierarchy_lookup(
    scope: ScopeDomain,
    registry: Optional[ScopeRegistry] = None,
    **kwargs,
) -> Set[ScopeDomain]:
    """Resolves scope hierarchy from the registry."""

    if registry is None:
        return {scope}

    found = registry.lookup_hierarchy(scope)

    if found is None:
        found = registry.lookup_hierarchy(scope.get_basement())

    return {scope} | (set(found.lookup_ancestors()) if found is not None else set())


def domain_path_hierarchy_lookup(
    scope: ScopeDomain, **kwargs
) -> Set[ScopeDomain]:
    """Resolves scope hierarchy defined in domain."""

    return set(scope.get_path())


class HierarchyChecker(ScopesChecker):
    """Hierarchical scopes checker.

    It finds all scope hierarchy and runs internal checkers for all
    scopes it found.
    """
    hierarchy_lookup: Callable
    checkers: Sequence[ScopesCheckerType]

    def __init__(
        self,
        checkers: Sequence[ScopesCheckerType],
        hierarchy_lookup: Callable = registry_hierarchy_lookup
    ):
        self.checkers = checkers
        self.hierarchy_lookup = hierarchy_lookup

    def has(
        self,
        check_scopes: Set[ScopeDomain],
        **kwargs
    ) -> bool:
        resolved_scopes = reduce(
            lambda acc, x: acc | self.hierarchy_lookup(x, **kwargs),
            check_scopes,
            set()
        )

        return run_checkers(resolved_scopes, self.checkers, kwargs)



from functools import lru_cache


__all__ = (
    'get_permission_app_label',
    'get_permission_string',
)


@lru_cache
def get_permission_app_label() -> str:
    from .models import Scope

    return Scope._meta.app_label


def get_permission_string(key: str) -> str:
    return get_permission_app_label() + '.' + key



from typing import Iterable, Iterator, Set, Type, TypeVar, Union, TYPE_CHECKING
from enum import Enum

from .aggregates import Aggregate, Aggregates

if TYPE_CHECKING:
    from .scopes import ScopeDomain


T = TypeVar('T')
D = TypeVar('D', bound='ScopeDomain')
ScopeType = Union['ScopeDomain', Enum, str]


def to_set(value: Iterable[T]) -> Set[T]:
    return value if isinstance(value, set) else set(value)


def normalize_scope(scope: ScopeType, cls: Type[D] = 'ScopeDomain') -> D:
    if isinstance(scope, Enum):
        scope = scope.value

    if not isinstance(scope, cls):
        scope = cls.parse(scope)

    return scope


def flatten_aggregates(
    aggregates: Iterable[Union['Aggregate', 'Aggregates']]
) -> Iterator['Aggregate']:
    return (
        a
        for agg in aggregates
        for a in (
            (getattr(agg, name) for name in agg._item_names)
            if isinstance(agg, type) and issubclass(agg, Aggregates)
            else (agg,)
        )
    )



from px_access_scopes import export_nested_tree, nested_tree_to_leaves

from .globals import registries


__all__ = 'export_scopes',


def export_scopes(as_leaves: bool = False):
    tree = {}

    for root in registries:
        tree.update(export_nested_tree(root))

    return tree if not as_leaves else nested_tree_to_leaves(tree)



from typing import Set

from ..scopes import ScopeDomain
from .mixins import HasCallableMixin


__all__ = 'ScopesChecker',


class ScopesChecker(HasCallableMixin):
    def has(
        self,
        check_scopes: Set[ScopeDomain],
        *_,
        **kwargs
    ) -> bool:
        raise NotImplemented('Implement `has` method. It\'s mandatory.')



from .generator import generate_all
from .globals import registries, aggregates


__all__ = 'generate_data_on_migrate',


def generate_data_on_migrate(sender, **kwargs):
    generate_all(registries, aggregates)



from typing import Dict, Optional, Sequence
from functools import lru_cache, reduce
from django.http import Http404
from rest_framework.permissions import BasePermission, SAFE_METHODS

from px_access_scopes import (
    ScopesChecker, HierarchyChecker, ScopesCheckRunner,
    MultiregistryHierarchyLookup
)
from px_access_scopes.contrib.django import ScopeDomain, user_checker
from px_access_scopes.contrib.django.globals import registries


__all__ = 'ScopePermission', 'ScopeObjectPermission'


class ScopePermission(BasePermission):
    _METHODS: Sequence[str] = {
        'GET', 'OPTIONS', 'HEAD',
        'POST', 'PUT', 'PATCH', 'DELETE',
    }
    _AUTO_METHODS: Dict[str, Sequence[str]] = {
        'GET': ('OPTIONS', 'HEAD',),
        'POST': ('OPTIONS', 'HEAD',),
        'PUT': ('OPTIONS', 'HEAD',),
        'PATCH': ('OPTIONS', 'HEAD',),
        'DELETE': ('OPTIONS', 'HEAD',),
    }
    permissions_map: Dict[str, Sequence[str]] = {}
    checker: ScopesChecker = ScopesCheckRunner((
        HierarchyChecker(
            (user_checker,),
            hierarchy_lookup=MultiregistryHierarchyLookup(
                registries=registries
            )
        ),
    ))

    def get_required_permissions(self, method):
        """
        Return the list of permission codes that the user is required to have.
        """

        return self.permissions_map.get(method, [])

    def has_permission(self, request, view):
        if not request.user:
            return False

        return self.checker(
            self.get_required_permissions(request.method),
            user=request.user, request=request, view=view, method=request.method,
        )

    @classmethod
    @lru_cache(maxsize=600)
    def from_permissions(
        cls,
        *permissions: Sequence[ScopeDomain],
        methods: Optional[Sequence[str]] = None
    ) -> 'ScopePermission':
        if methods is None:
            methods = cls._METHODS
        auto_methods = cls._AUTO_METHODS

        assert all(False for x in methods if x not in cls._METHODS), (
            f'Unhandelable methods provided: {methods} \r\n'
            f'Must be one of: {cls._METHODS}.'
        )

        def permissions_map_reducer(map, method):
            map[method] = permissions

            for m in auto_methods.get(method, []):
                if m not in map:
                    map[m] = permissions

            return map

        permissions_map = reduce(permissions_map_reducer, methods, {})

        return type(cls.__name__, (cls, ), {
            'permissions_map': permissions_map
        })

    @classmethod
    @lru_cache(maxsize=600)
    def from_scopes(
        cls,
        *scopes: Sequence[ScopeDomain],
        methods: Optional[Sequence[str]] = None
    ) -> 'ScopePermission':
        return cls.from_permissions(
            *(scope.permission for scope in scopes), methods=methods
        )


class ScopeObjectPermission(ScopePermission):
    def get_required_object_permissions(self, method):
        """
        Return the list of permission codes that the user is required to have.
        """

        return self.permissions_map.get(method, [])

    def has_object_permission(self, request, view, obj):
        kw = {
            'request': request,
            'view': view,
            'method': request.method,
            'user': request.user,
            'obj': obj
        }

        perms = self.get_required_object_permissions(request.method)

        if not self.checker(perms, **kw):
            # If the user does not have permissions we need to determine if
            # they have read permissions to see 403, or not, and simply see
            # a 404 response.

            if request.method in SAFE_METHODS:
                # Read permissions already checked and failed, no need
                # to make another lookup.
                raise Http404

            read_perms = self.get_required_object_permissions('GET')
            if not self.checker(read_perms, **kw):
                raise Http404

            # Has read permissions.
            return False

        return True



from enum import Enum
from px_domains import Domain, DomainRegistry, raw


__all__ = (
    'auto',
    'raw',
    'ScopeDomain',
    'ScopeEnum',
    'ScopeRegistry',
)


class ScopeDomain(Domain):
    pass


class ScopeEnum(ScopeDomain, Enum):
    pass


auto = ScopeDomain('auto')


class ScopeRegistry(DomainRegistry):
    """Simple domain registry for an access scopes.

    Examples:
        Creating registry root

        >>> root = ScopeRegistry.create_root('ROOT')
        >>>
        >>> @root.nest('SOME')
        >>> class Tokens(Enum):
        >>>     AUTO1 = auto
        >>>     AUTO2 = auto
        >>>     AUTO3 = raw('AUTO3')
        >>>     FIXED = 'some'

        Can be nested in every possible way.

        >>> @Tokens.nest('NESTED')
        >>> class Nested(Enum):
        >>>     AUTO = auto

        `Tokens.AUTO1` will be: `'ROOT::SOME::AUTO1'`
    """

    _Domain = ScopeDomain
    _Enum = ScopeEnum
    _auto = auto



from collections import deque

from .scopes import ScopeRegistry


__all__ = (
    'export_nested_tree',
    'nested_tree_to_leaves',
    'export_nested_tree_leaves',
)


VALUE_KEY = 'v'
CHILDREN_KEY = 'c'


def export_nested_tree(
    registry: ScopeRegistry,
    value_key: str = VALUE_KEY,
    children_key: str = CHILDREN_KEY
) -> dict:
    current = registry._domain
    maps = {}
    state = deque(((None, current, current._definition[-1]),))
    h = registry._hierarchy
    s = registry._sets

    while True:
        try:
            parent, current, key = state.popleft()
        except IndexError:
            break

        maps[parent] = maps.get(parent, {
            value_key: parent, children_key: {},
        })
        maps[current] = {
            value_key: current, children_key: {},
        }
        maps[parent][children_key][key] = maps[current]

        if current in h:
            for scope in h[current]:
                state.append((current, scope, scope._definition[-1]))

        if current in s:
            for e in s[current]:
                state.append((current, e, e._name_))

    return maps[None][children_key]


def nested_tree_to_leaves(
    tree: dict,
    value_key: str = VALUE_KEY,
    children_key: str = CHILDREN_KEY
) -> dict:
    return {
        key: (
            values[value_key]
            if len(values[children_key].keys()) == 0
            else nested_tree_to_leaves(values[children_key])
        )
        for key, values in tree.items()
    }


def export_nested_tree_leaves(registry: ScopeRegistry) -> dict:
    exported = export_nested_tree(registry)

    return nested_tree_to_leaves(exported)
